//////////////////////////////////////////////////////////////////////////
// spieleprogrammierer.de-Programmier-Contest #08                       //
// ******************************************************************** //
// Titel: ...... "Pong"                                                 //
// Typ: ........ Spezial                                                //
// Deadline: ... 07.09.2008                                             //
// Abgabe: ..... contest@spieleprogrammierer.de                         //
// ******************************************************************** //
//                                                                      //
// Aufgabenstellung siehe:                                              //
// http://www.spieleprogrammierer.de/phpBB2/viewtopic.php?t=10062       //
//                                                                      //
//////////////////////////////////////////////////////////////////////////

#include <vector>
#include <iostream>
#include <ctime>
#include <cmath>
#include <cfloat>
#include "pong.h"

//////////////////////////////////////////////////////////////////////////

// Diese Konstanten werden sich auch bei der Auswertung nicht ndern.
// Ihr knnt sie also benutzen oder direkt mit den Zahlwerten arbeiten.

// Breite des Spielfelds
const float FIELD_WIDTH = 2.0f;

// Hhe des Spielfelds
const float FIELD_HEIGHT = 1.0f;

// Lnge der Schlger
const float PADDLE_SIZE = 0.1f;

// Beschleunigung des Spielers
const float PLAYER_ACCELERATION = 0.01f;

// Dmpfung der Spielergeschwindigkeit
const float PLAYER_DAMPING = 0.75f;

// Startgeschwindigkeit des Balls
const float START_BALL_SPEED = 0.05f;

// Erhhung der Ballgeschwindigkeit mit jedem Schlag (Faktor)
const float BALL_SPEED_INCREASE = 1.01f;

//////////////////////////////////////////////////////////////////////////

// Mersenne Twister-Zufallsgenerator
class MTRand
{
public:
    MTRand(uint seed = 0) { this->seed(seed); }

    void seed(uint seed)
    {
        next = 0;
        const unsigned long long int factor = 69069;
        mt[0] = seed;
        for(uint i = 1; i != 624; ++i) mt[i] = static_cast<uint>(factor * mt[i - 1] + 1);
    }

    float float_0_to_1()
    {
        const uint x = (*this)() & 0xFFFFFF;
        return static_cast<float>(x) / 0xFFFFFF;
    }

    uint operator ()()
    {
        if(next == 624) { generateNumbers(); next = 0; }
        uint x = mt[next++];
        x ^= x >> 11;
        x ^= (x << 7) & 0x9D2C5680;
        x ^= (x << 15) & 0xEFC60000;
        x ^= x >> 18;
        return x;
    }

private:
    void generateNumbers()
    {
        const uint xorTab[] = { 0, 0x9908B0DF };
        for(uint i = 0; i != 227; ++i) { const uint x = (mt[i] & 0x80000000) | (mt[i + 1] & 0x7FFFFFFF); mt[i] = mt[i + 397] ^ (x >> 1) ^ xorTab[x & 1]; }
        for(uint i = 227; i != 624; ++i) { const uint x = (mt[i] & 0x80000000) | (mt[i + 1] & 0x7FFFFFFF); mt[i] = mt[i - 227] ^ (x >> 1) ^ xorTab[x & 1]; }
        const uint x = (mt[623] & 0x80000000) | (mt[0] & 0x7FFFFFFF);
        mt[623] = mt[396] ^ (x >> 1) ^ xorTab[x & 1];
    }

    uint mt[624];
    uint next;
};

//////////////////////////////////////////////////////////////////////////

// >>> DEINE FUNKTION - BITTE IMPLEMENTIEREN! <<<
int my_pong(const GameState& gs,    // Spielzustand
            uint me)                // Welcher Spieler bin ich? (0 oder 1)
{
    // Rckgabewert:
    // -1: nach oben bewegen
    //  0: nicht bewegen
    //  1: nach unten bewegen
    return 0;
}

//////////////////////////////////////////////////////////////////////////

// Referenzfunktion
int reference_pong(const GameState& gs,
                   uint me)
{
    if(gs.ball_y < gs.player_y[me] - 0.25f * PADDLE_SIZE) return -1;
    else if(gs.ball_y > gs.player_y[me] + 0.25f * PADDLE_SIZE) return 1;
    else return 0;
}

//////////////////////////////////////////////////////////////////////////

// fhrt ein einzelnes Spiel zwischen zwei Spielern durch
void do_game(const Entry& player_1,
             const Entry& player_2,
             uint time_limit,
             uint score_limit,
             bool visualize,
             uint& out_score_1,
             uint& out_score_2,
             uint& out_time_played)
{
    enum Event
    {
        E_NOTHING,
        E_REACH_LEFT_GOAL_LINE,
        E_REACH_RIGHT_GOAL_LINE,
        E_REACH_TOP_WALL,
        E_REACH_BOTTOM_WALL
    };

    MTRand rnd(time(0));

    float ball_speed = START_BALL_SPEED;
    float real_time = 0.0f;
    uint last_time = ~0;
    int player_action[2] = { 0, 0 };

    GameState gs;
    gs.p_name[0] = player_1.p_name;
    gs.p_name[1] = player_2.p_name;
    gs.score[0] = gs.score[1] = 0;
    gs.time = 0;
    gs.player_y[0] = gs.player_y[1] = 0.5f * FIELD_HEIGHT;
    gs.player_vy[0] = gs.player_vy[1] = 0.0f;
    gs.ball_x = 0.5f * FIELD_WIDTH;
    gs.ball_y = 0.5f * FIELD_HEIGHT;
    gs.ball_vx = ball_speed * (rnd() % 2 ? 1.0f : -1.0f);
    gs.ball_vy = ball_speed * (rnd.float_0_to_1() - 0.5f);

    while(gs.time < time_limit &&
          gs.score[0] < score_limit &&
          gs.score[1] < score_limit)
    {
        // Die KIs werden einmal pro Frame aufgerufen.
        gs.time = static_cast<uint>(real_time);
        if(last_time != gs.time)
        {
            last_time = gs.time;
            player_action[0] = player_1.function(gs, 0);
            player_action[1] = player_2.function(gs, 1);
        }

        // Wann berhrt der Ball zum nchsten Mal eine Bande oder die Torlinie?
        // Diese Zeit kann "sicher" in einem einzigen Schritt simuliert werden.
        // Wir beschrnken die Zeit aber auf ein Frame.
        Event what_happens = E_NOTHING;
        float safe_dt = static_cast<float>(gs.time + 1) - real_time;
        if(gs.ball_vx < 0.0f && gs.ball_x / -gs.ball_vx < safe_dt) safe_dt = gs.ball_x / -gs.ball_vx, what_happens = E_REACH_LEFT_GOAL_LINE;
        else if(gs.ball_vx > 0.0f && (FIELD_WIDTH - gs.ball_x) / gs.ball_vx < safe_dt) safe_dt = (FIELD_WIDTH - gs.ball_x) / gs.ball_vx, what_happens = E_REACH_RIGHT_GOAL_LINE;
        if(gs.ball_vy < 0.0f && gs.ball_y / -gs.ball_vy < safe_dt) safe_dt = gs.ball_y / -gs.ball_vy, what_happens = E_REACH_TOP_WALL;
        else if(gs.ball_vy > 0.0f && (FIELD_HEIGHT - gs.ball_y) / gs.ball_vy < safe_dt) safe_dt = (FIELD_HEIGHT - gs.ball_y) / gs.ball_vy, what_happens = E_REACH_BOTTOM_WALL;

        // Spieler bewegen
        for(uint i = 0; i < 2; ++i)
        {
            if(player_action[i] == -1) gs.player_vy[i] -= safe_dt * PLAYER_ACCELERATION;
            else if(player_action[i] == 1) gs.player_vy[i] += safe_dt * PLAYER_ACCELERATION;
            gs.player_vy[i] *= powf(PLAYER_DAMPING, safe_dt);
            gs.player_y[i] += safe_dt * gs.player_vy[i];

            // Position auf das Spielfeld beschrnken
            if(gs.player_y[i] < 0.5f * PADDLE_SIZE) gs.player_y[i] = 0.5f * PADDLE_SIZE, gs.player_vy[i] = 0.0f;
            else if(gs.player_y[i] > FIELD_HEIGHT - 0.5f * PADDLE_SIZE) gs.player_y[i] = FIELD_HEIGHT - 0.5f * PADDLE_SIZE, gs.player_vy[i] = 0.0f;
        }

        int player_scored = -1;

        // Ball bewegen
        switch(what_happens)
        {
            case E_NOTHING:
                gs.ball_x += safe_dt * gs.ball_vx;
                gs.ball_y += safe_dt * gs.ball_vy;
                break;

            case E_REACH_LEFT_GOAL_LINE:
                gs.ball_x = 0.0f;
                gs.ball_y += safe_dt * gs.ball_vy;

                // auf Tor berprfen
                if(gs.ball_y >= gs.player_y[0] - 0.5f * PADDLE_SIZE &&
                   gs.ball_y <= gs.player_y[0] + 0.5f * PADDLE_SIZE)
                {
                    // Gehalten!
                    gs.ball_vx = -gs.ball_vx;
                    gs.ball_vy -= gs.player_vy[0] * (0.95f + 0.1f * rnd.float_0_to_1());
                    gs.ball_vx *= BALL_SPEED_INCREASE;
                    gs.ball_vy *= BALL_SPEED_INCREASE;
                }
                else
                {
                    // Tor!
                    player_scored = 1;
                }
                break;

            case E_REACH_RIGHT_GOAL_LINE:
                gs.ball_x = FIELD_WIDTH;
                gs.ball_y += safe_dt * gs.ball_vy;

                // auf Tor berprfen
                if(gs.ball_y >= gs.player_y[1] - 0.5f * PADDLE_SIZE &&
                   gs.ball_y <= gs.player_y[1] + 0.5f * PADDLE_SIZE)
                {
                    gs.ball_vx = -gs.ball_vx;
                    gs.ball_vy -= gs.player_vy[1] * (0.95f + 0.1f * rnd.float_0_to_1());
                    gs.ball_vx *= BALL_SPEED_INCREASE;
                    gs.ball_vy *= BALL_SPEED_INCREASE;
                }
                else
                {
                    // Tor!
                    player_scored = 0;
                }
                break;

            case E_REACH_TOP_WALL:
                // prallt von der oberen Bande ab
                gs.ball_x += safe_dt * gs.ball_vx;
                gs.ball_y = 0.0f;
                gs.ball_vy = -gs.ball_vy;
                break;

            case E_REACH_BOTTOM_WALL:
                // prallt von der unteren Bande ab
                gs.ball_x += safe_dt * gs.ball_vx;
                gs.ball_y = FIELD_HEIGHT;
                gs.ball_vy = -gs.ball_vy;
                break;
        }

        // Wenn ein Punkt gemacht wurde, geht's zurck in die Ausgangslage,
        // und der andere Spieler bekommt "Ansto". Auerdem wird der Ball schneller.
        if(player_scored != -1)
        {
            ++gs.score[player_scored];
            ball_speed = START_BALL_SPEED;
            gs.player_y[0] = gs.player_y[1] = 0.5f * FIELD_HEIGHT;
            gs.player_vy[0] = gs.player_vy[1] = 0.0f;
            gs.ball_x = 0.5f * FIELD_WIDTH;
            gs.ball_y = 0.5f * FIELD_HEIGHT;
            gs.ball_vx = ball_speed * (player_scored ? 1.0f : -1.0f);
            gs.ball_vy = ball_speed * (rnd.float_0_to_1() - 0.5f);
        }

        if(visualize)
        {
            // Visualisierung aktualisieren
            visualizer_update(gs);
        }

        real_time += safe_dt;
    }

    // Ergebnis eintragen
    out_score_1 = gs.score[0];
    out_score_2 = gs.score[1];
    out_time_played = gs.time;
}

//////////////////////////////////////////////////////////////////////////

int main()
{
    visualizer_start();

    Entry reference("Referenz", &reference_pong);
    Entry me("Ich", &my_pong);

    uint score_1, score_2, time_played;
    do_game(reference, me, ~0, 100, true, score_1, score_2, time_played);
    std::cout << reference.p_name << ' ' << score_1 << ':' << score_2 << ' ' << me.p_name << " (" << time_played << ')' << std::endl;

    visualizer_end();

    return 0;
}
