// Pong-Visualisierung fr Windows

#include <sstream>
#include <windows.h>
#include "pong.h"

WNDCLASS wc;
HWND wnd;
bool window_open;
GameState draw_gs;
bool draw_gs_valid;
HDC buffer_dc;
uint delay = 30;
HANDLE message_thread;

LRESULT CALLBACK window_proc(HWND wnd,
                             UINT msg,
                             WPARAM wp,
                             LPARAM lp)
{
    switch(msg)
    {
    case WM_CREATE:
        window_open = true;
        break;
    case WM_CLOSE:
        window_open = false;
        DestroyWindow(wnd);
        break;
    case WM_CHAR:
        if(LOWORD(wp) == '-' && delay < 100) delay += 5;
        else if(LOWORD(wp) == '+' && delay) delay -= 5;
        break;
    case WM_DESTROY:
        PostQuitMessage(0);
        break;
    case WM_PAINT:
        if(draw_gs_valid)
        {
            RECT rect;
            SetRect(&rect, 0, 0, 600, 300);
            FillRect(buffer_dc, &rect, reinterpret_cast<HBRUSH>(GetStockObject(BLACK_BRUSH)));
            SetRect(&rect, 0, static_cast<int>((draw_gs.player_y[0] - 0.5f * PADDLE_SIZE) * 300.0f), 6, static_cast<int>((draw_gs.player_y[0] + 0.5f * PADDLE_SIZE) * 300.0f));
            FillRect(buffer_dc, &rect, reinterpret_cast<HBRUSH>(GetStockObject(WHITE_BRUSH)));
            SetRect(&rect, 594, static_cast<int>((draw_gs.player_y[1] - 0.5f * PADDLE_SIZE) * 300.0f), 600, static_cast<int>((draw_gs.player_y[1] + 0.5f * PADDLE_SIZE) * 300.0f));
            FillRect(buffer_dc, &rect, reinterpret_cast<HBRUSH>(GetStockObject(WHITE_BRUSH)));
            SetRect(&rect, 6 + static_cast<int>(draw_gs.ball_x * 294.0f - 2.5f), static_cast<int>(draw_gs.ball_y * 300.0f - 2.5f), 6 + static_cast<int>(draw_gs.ball_x * 294.0f + 2.5f), static_cast<int>(draw_gs.ball_y * 300.0f + 2.5f));
            FillRect(buffer_dc, &rect, reinterpret_cast<HBRUSH>(GetStockObject(LTGRAY_BRUSH)));
            SetRect(&rect, 0, 300, 600, 340);
            FillRect(buffer_dc, &rect, reinterpret_cast<HBRUSH>(GetStockObject(DKGRAY_BRUSH)));
            std::ostringstream str;
            str << draw_gs.p_name[0] << ' ' << draw_gs.score[0] << " : " << draw_gs.score[1] << ' ' << draw_gs.p_name[1] << std::endl << "Zeit: " << draw_gs.time;
            SetBkMode(buffer_dc, TRANSPARENT);
            SetTextColor(buffer_dc, RGB(255, 255, 255));
            DrawText(buffer_dc, str.str().c_str(), -1, &rect, DT_CENTER | DT_VCENTER);

            PAINTSTRUCT ps;
            HDC dc = BeginPaint(wnd, &ps);
            BitBlt(dc, ps.rcPaint.left, ps.rcPaint.top, ps.rcPaint.right - ps.rcPaint.left, ps.rcPaint.bottom - ps.rcPaint.top, buffer_dc, ps.rcPaint.left, ps.rcPaint.top, SRCCOPY);
            EndPaint(wnd, &ps);
        }
        break;
    default:
        return DefWindowProc(wnd, msg, wp, lp);
    }

    return TRUE;
}

DWORD WINAPI thread_proc(void* p_param)
{
    wc.style = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc = &window_proc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
    wc.hInstance = GetModuleHandle(0);
    wc.hCursor = LoadCursor(0, IDC_ARROW);
    wc.hIcon = LoadIcon(0, IDI_APPLICATION);
    wc.hbrBackground = 0;
    wc.lpszClassName = "spieleprogrammierer.de-Programmier-Contest #08: Pong";
    wc.lpszMenuName = 0;
    RegisterClass(&wc);

    RECT rect;
	SetRect(&rect, 0, 0, 600, 340);
	AdjustWindowRect(&rect, WS_VISIBLE | WS_POPUPWINDOW | WS_MINIMIZEBOX | WS_CAPTION, FALSE);

    wnd = CreateWindow(wc.lpszClassName,
                       wc.lpszClassName,
                       WS_VISIBLE | WS_POPUPWINDOW | WS_MINIMIZEBOX | WS_CAPTION,
                       100,
                       100,
                       rect.right - rect.left,
                       rect.bottom - rect.top,
                       0, 0, GetModuleHandle(0), 0);

    HDC dc = GetDC(wnd);
    buffer_dc = CreateCompatibleDC(dc);
    HBITMAP buffer_bmp = CreateCompatibleBitmap(dc, 600, 340);
    ReleaseDC(wnd, dc);
    SelectObject(buffer_dc, buffer_bmp);

    MSG msg;
    while(GetMessage(&msg, 0, 0, 0))
    {
        if(msg.message == WM_QUIT) break;

        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }

    DeleteDC(buffer_dc);
    DeleteObject(buffer_bmp);

    UnregisterClass(wc.lpszClassName, GetModuleHandle(0));

    return 0;
}

void visualizer_start()
{
    message_thread = CreateThread(0, 0, &thread_proc, 0, 0, 0);
    while(!window_open) Sleep(0);
}

void visualizer_update(const GameState& gs)
{
    if(!window_open) return;

    draw_gs = gs;
    draw_gs_valid = true;
    InvalidateRect(wnd, 0, TRUE);
    UpdateWindow(wnd);
    Sleep(delay);
}

void visualizer_end()
{
    if(window_open) SendMessage(wnd, WM_CLOSE, 0, 0);
    WaitForSingleObject(message_thread, INFINITE);
}
