#include <sstream>
#include <unistd.h>
#include <pthread.h>
#include <X11/Xlib.h>
#include <X11/keysym.h>
#include "pong.h"

Display* p_display;
int screenNo;
Window window;
bool window_open = true;
Pixmap buffer;
GC window_gc;
GC buffer_gc;
Font font;
uint delay = 30;
pthread_t message_thread;

void* thread_proc(void* p_arg)
{
    XEvent evt;
    XSelectInput(p_display, window, StructureNotifyMask | KeyPressMask);
    bool quit = false;

    do
    {
        XNextEvent(p_display, &evt);
        switch(evt.type)
        {
        case DestroyNotify:
            quit = true;
            break;
        case KeyPress:
            if(evt.xkey.keycode == XKeysymToKeycode(p_display, XK_KP_Subtract) && delay < 100) delay += 5;
            else if(evt.xkey.keycode == XKeysymToKeycode(p_display, XK_KP_Add) && delay) delay -= 5;
            break;
        case ClientMessage:
            if(evt.xclient.message_type == XInternAtom(p_display, "WM_PROTOCOLS", 0) && evt.xclient.data.l[0] == XInternAtom(p_display, "WM_DELETE_WINDOW", 0)) quit = true;
            break;
        }
    }
    while(!quit);

    XUnmapWindow(p_display, window);
    window_open = false;
}

void visualizer_start()
{
    p_display = XOpenDisplay(0);
    screenNo = DefaultScreen(p_display);

    window = XCreateSimpleWindow(p_display, DefaultRootWindow(p_display),
                                 100, 100, 600, 340,
                                 0, 0, 0);

    Atom delete_window = XInternAtom(p_display, "WM_DELETE_WINDOW", 0);
    XSetWMProtocols(p_display, window, &delete_window, 1);

    XMapWindow(p_display, window);

    buffer = XCreatePixmap(p_display, window, 600, 340, DefaultDepth(p_display, screenNo));
    font = XLoadFont(p_display, "9x15");

    XSelectInput(p_display, window, StructureNotifyMask);
    XEvent evt;
    do XNextEvent(p_display, &evt); while(evt.type != MapNotify);

    window_gc = XCreateGC(p_display, window, 0, 0);
    buffer_gc = XCreateGC(p_display, buffer, 0, 0);

    pthread_create(&message_thread, 0, &thread_proc, 0);
}

void visualizer_update(const GameState& gs)
{
    if(!window_open) return;

    clock_t t0 = clock();

    XSetForeground(p_display, buffer_gc, 0);
    XFillRectangle(p_display, buffer, buffer_gc, 0, 0, 600, 340);

    XSetForeground(p_display, buffer_gc, 0xFFFFFF);
    XFillRectangle(p_display, buffer, buffer_gc, 0, static_cast<int>((gs.player_y[0] - 0.5f * PADDLE_SIZE) * 300.0f), 6, static_cast<int>(PADDLE_SIZE * 300.0f));
    XFillRectangle(p_display, buffer, buffer_gc, 594, static_cast<int>((gs.player_y[1] - 0.5f * PADDLE_SIZE) * 300.0f), 6, static_cast<int>(PADDLE_SIZE * 300.0f));

    XSetForeground(p_display, buffer_gc, 0xA0A0A0);
    XFillRectangle(p_display, buffer, buffer_gc, 6 + static_cast<int>(gs.ball_x * 294.0f - 2.5f), static_cast<int>(gs.ball_y * 300.0f - 2.5f), 5, 5);

    XSetForeground(p_display, buffer_gc, 0x808080);
    XFillRectangle(p_display, buffer, buffer_gc, 0, 300, 600, 40);

    std::ostringstream str;
    str << gs.p_name[0] << ' ' << gs.score[0] << " : " << gs.score[1] << ' ' << gs.p_name[1] << " (Zeit: " << gs.time << ')';

    XSetForeground(p_display, buffer_gc, 0xFFFFFF);
    XSetFont(p_display, buffer_gc, font);
    XTextItem text;
    text.chars = const_cast<char*>(str.str().c_str());
    text.nchars = str.str().length();
    text.delta = 0;
    text.font = 0;
    XDrawText(p_display, buffer, buffer_gc, 20, 320, &text, 1);

    XCopyArea(p_display, buffer, window, window_gc, 0, 0, 600, 340, 0, 0);
    XFlush(p_display);

    while(clock() < static_cast<clock_t>(CLOCKS_PER_SEC * delay / 1000 + t0)) sched_yield();
}

void visualizer_end()
{
    XDestroyWindow(p_display, window);
    pthread_join(message_thread, 0);

    XUnloadFont(p_display, font);
    XFreePixmap(p_display, buffer);
    XCloseDisplay(p_display);
}
